#!/usr/bin/env python

import po
import sys
from pprint import pprint
import difflib
import textwrap

def highlight_nonmatching_blocks(matching_blocks, which, string):
	previous_match_end = 0
	highlight_string = ""
	for block in matching_blocks:
		(begin_match_1, begin_match_2, match_len) = block
		
		if which == 1:
			begin_match = begin_match_1
		else:
			begin_match = begin_match_2
			
		#print "diff_len=" + str(begin_match - previous_match_end), "match_len=" + str(match_len)
		
		highlight_string = highlight_string + ''.rjust(begin_match - previous_match_end, '^')
		previous_match_end = begin_match + match_len
		highlight_string = highlight_string + ''.rjust(match_len, ' ')
		
	i = 0
	result = ""
	for line in string.split("\n"):
		result = result + line + "\n"
		result = result + highlight_string[i:i+len(line)] + "\n"
		i = i + len(line) + 1
	return result
	
def highlight_strings_differences_ascii(s1, s2):
	d = difflib.SequenceMatcher()
	d.set_seqs(s1, s2)
	matching_blocks = d.get_matching_blocks()
	result1 = highlight_nonmatching_blocks(matching_blocks, 1, s1)
	result2 = highlight_nonmatching_blocks(matching_blocks, 2, s2)
	return (result1, result2)

def print_entries(keysToPrint, entries):
	for key in keysToPrint:
		poentry = entries[key]
		print "msgid=" + poentry["msgid"].encode('utf-8')
		print "msgstr=" + poentry["msgstr"].encode('utf-8')

def print_differing_entries(keysToPrint, entries1, entries2, orderedEntries):
	#walk orderedEntries so that order is kept when printing
	for entry in orderedEntries:
		key = entry['msgid']
		if key not in keysToPrint:
			continue
		poentry = entries1[key]
		poentry2 = entries2[key]
		wrapLen = 80
		msgid = textwrap.fill(poentry["msgid"], wrapLen)
		msgstr1 = textwrap.fill(poentry["msgstr"], wrapLen)
		msgstr2 = textwrap.fill(poentry2["msgstr"], wrapLen)
		print ''.rjust(wrapLen, "#")
		print "msgid:"
		print msgid.encode('utf-8')
		print
		#print "msgstr1=" + poentry["msgstr"].encode('utf-8')
		#print "msgstr2=" + poentry2["msgstr"].encode('utf-8')
		(diff1, diff2) = highlight_strings_differences_ascii(msgstr1, msgstr2)
		
		if poentry["fuzzy"] == True: fuzzy1=" [fuzzy]" 
		else: fuzzy1=""
		if poentry2["fuzzy"] == True: fuzzy2=" [fuzzy]" 
		else: fuzzy2=""
		
		print "msgstr1" + fuzzy1 + ":"
		print diff1.encode('utf-8')
		print "msgstr2" + fuzzy2 + ":"
		print diff2.encode('utf-8')

def usage():
	print "Tool comparing .po files based on msgids."
	print "Differences in msgstrs are highlighted."
	print "Usage: msgtool file1.po file2.po"
		
def main():
	if len(sys.argv) != 3:
		usage()
		sys.exit(1)
	
	filename1 = sys.argv[1]
	filename2 = sys.argv[2]
	(entries1, keys1) = po.read(filename1)
	(entries2, keys2) = po.read(filename2)
	keysset1 = set(keys1.keys())
	keysset2 = set(keys2.keys())
	
	only_in_1 = keysset1.difference(keysset2)
	only_in_2 = keysset2.difference(keysset1)
	common = keysset2.intersection(keysset1)
	
	print
	print("********* Only in 1: ************\n")
	print_entries(only_in_1, keys1)
	
	print
	print("********* Only in 2: ************\n")
	print_entries(only_in_2, keys2)
	
	differentTranslations = set()
	
	print
	print("********* Different translations: ************\n")
	for key in common:
		if (keys1[key]["msgstr"] != keys2[key]["msgstr"]) or (keys1[key]["fuzzy"] != keys2[key]["fuzzy"]):
			differentTranslations.add(key)
			
	print_differing_entries(differentTranslations, keys1, keys2, entries1)

main()

